@@ -3,14 +3,38 @@
 # Copyright (c) PyZMQ Developers.
 # Distributed under the terms of the Modified BSD License.
 
-from collections import namedtuple, deque
+import warnings
+from asyncio import Future
+from collections import deque
 from itertools import chain
-from typing import Type
+from typing import (
+    Any,
+    Awaitable,
+    Callable,
+    Dict,
+    List,
+    NamedTuple,
+    Optional,
+    Tuple,
+    Type,
+    TypeVar,
+    Union,
+    cast,
+    overload,
+)
 
-from zmq import EVENTS, POLLOUT, POLLIN
 import zmq as _zmq
+from zmq import EVENTS, POLLIN, POLLOUT
+from zmq._typing import Literal
+
+
+class _FutureEvent(NamedTuple):
+    future: Future
+    kind: str
+    kwargs: Dict
+    msg: Any
+    timer: Any
 
-_FutureEvent = namedtuple('_FutureEvent', ('future', 'kind', 'kwargs', 'msg'))
 
 # These are incomplete classes and need a Mixin for compatibility with an eventloop
 # defining the following attributes:
@@ -21,30 +45,71 @@ _FutureEvent = namedtuple('_FutureEvent', ('future', 'kind', 'kwargs', 'msg'))
 # _default_loop()
 
 
-class _AsyncPoller(_zmq.Poller):
+class _Async:
+    """Mixin for common async logic"""
+
+    _current_loop: Any = None
+    _Future: Type[Future]
+
+    def _get_loop(self) -> Any:
+        """Get event loop
+
+        Notice if event loop has changed,
+        and register init_io_state on activation of a new event loop
+        """
+        if self._current_loop is None:
+            self._current_loop = self._default_loop()
+            self._init_io_state(self._current_loop)
+            return self._current_loop
+        current_loop = self._default_loop()
+        if current_loop is not self._current_loop:
+            # warn? This means a socket is being used in multiple loops!
+            self._current_loop = current_loop
+            self._init_io_state(current_loop)
+        return current_loop
+
+    def _default_loop(self) -> Any:
+        raise NotImplementedError("Must be implemented in a subclass")
+
+    def _init_io_state(self, loop=None) -> None:
+        pass
+
+
+class _AsyncPoller(_Async, _zmq.Poller):
     """Poller that returns a Future on poll, instead of blocking."""
 
-    _socket_class = None  # type: Type[_AsyncSocket]
+    _socket_class: Type["_AsyncSocket"]
+    _READ: int
+    _WRITE: int
+    raw_sockets: List[Any]
 
-    def poll(self, timeout=-1):
+    def _watch_raw_socket(self, loop: Any, socket: Any, evt: int, f: Callable) -> None:
+        """Schedule callback for a raw socket"""
+        raise NotImplementedError()
+
+    def _unwatch_raw_sockets(self, loop: Any, *sockets: Any) -> None:
+        """Unschedule callback for a raw socket"""
+        raise NotImplementedError()
+
+    def poll(self, timeout=-1) -> Awaitable[List[Tuple[Any, int]]]:  # type: ignore
         """Return a Future for a poll event"""
         future = self._Future()
         if timeout == 0:
             try:
-                result = super(_AsyncPoller, self).poll(0)
+                result = super().poll(0)
             except Exception as e:
                 future.set_exception(e)
             else:
                 future.set_result(result)
             return future
 
-        loop = self._default_loop()
+        loop = self._get_loop()
 
         # register Future to be called as soon as any event is available on any socket
         watcher = self._Future()
 
         # watch raw sockets:
-        raw_sockets = []
+        raw_sockets: List[Any] = []
 
         def wake_raw(*args):
             if not watcher.done():
@@ -119,46 +184,66 @@ class _AsyncPoller(_zmq.Poller):
         return future
 
 
-class _AsyncSocket(_zmq.Socket):
+class _NoTimer:
+    @staticmethod
+    def cancel():
+        pass
+
+
+T = TypeVar("T", bound="_AsyncSocket")
+
+
+class _AsyncSocket(_Async, _zmq.Socket):
 
     # Warning : these class variables are only here to allow to call super().__setattr__.
     # They be overridden at instance initialization and not shared in the whole class
     _recv_futures = None
     _send_futures = None
     _state = 0
-    _shadow_sock = None
+    _shadow_sock: "_zmq.Socket"
     _poller_class = _AsyncPoller
-    io_loop = None
     _fd = None
 
-    def __init__(self, context=None, socket_type=-1, io_loop=None, **kwargs):
+    def __init__(
+        self,
+        context=None,
+        socket_type=-1,
+        io_loop=None,
+        _from_socket: Optional["_zmq.Socket"] = None,
+        **kwargs,
+    ) -> None:
         if isinstance(context, _zmq.Socket):
-            context, from_socket = (None, context)
+            context, _from_socket = (None, context)
+        if _from_socket is not None:
+            super().__init__(shadow=_from_socket.underlying)
+            self._shadow_sock = _from_socket
         else:
-            from_socket = kwargs.pop('_from_socket', None)
-        if from_socket is not None:
-            super(_AsyncSocket, self).__init__(shadow=from_socket.underlying)
-            self._shadow_sock = from_socket
-        else:
-            super(_AsyncSocket, self).__init__(context, socket_type, **kwargs)
+            super().__init__(context, socket_type, **kwargs)
             self._shadow_sock = _zmq.Socket.shadow(self.underlying)
-        self.io_loop = io_loop or self._default_loop()
+
+        if io_loop is not None:
+            warnings.warn(
+                f"{self.__class__.__name__}(io_loop) argument is deprecated in pyzmq 22.2."
+                " The currently active loop will always be used.",
+                DeprecationWarning,
+                stacklevel=3,
+            )
         self._recv_futures = deque()
         self._send_futures = deque()
         self._state = 0
         self._fd = self._shadow_sock.FD
-        self._init_io_state()
 
     @classmethod
-    def from_socket(cls, socket, io_loop=None):
+    def from_socket(cls: Type[T], socket: "_zmq.Socket", io_loop: Any = None) -> T:
         """Create an async socket from an existing Socket"""
         return cls(_from_socket=socket, io_loop=io_loop)
 
-    def close(self, linger=None):
+    def close(self, linger: Optional[int] = None) -> None:
         if not self.closed and self._fd is not None:
-            for event in list(
+            event_list: List[_FutureEvent] = list(
                 chain(self._recv_futures or [], self._send_futures or [])
-            ):
+            )
+            for event in event_list:
                 if not event.future.done():
                     try:
                         event.future.cancel()
@@ -166,19 +251,45 @@ class _AsyncSocket(_zmq.Socket):
                         # RuntimeError may be called during teardown
                         pass
             self._clear_io_state()
-        super(_AsyncSocket, self).close(linger=linger)
+        super().close(linger=linger)
 
     close.__doc__ = _zmq.Socket.close.__doc__
 
     def get(self, key):
-        result = super(_AsyncSocket, self).get(key)
+        result = super().get(key)
         if key == EVENTS:
             self._schedule_remaining_events(result)
         return result
 
     get.__doc__ = _zmq.Socket.get.__doc__
 
-    def recv_multipart(self, flags=0, copy=True, track=False):
+    @overload  # type: ignore
+    def recv_multipart(
+        self, flags: int = 0, *, track: bool = False
+    ) -> Awaitable[List[bytes]]:
+        ...
+
+    @overload
+    def recv_multipart(
+        self, flags: int = 0, *, copy: Literal[True], track: bool = False
+    ) -> Awaitable[List[bytes]]:
+        ...
+
+    @overload
+    def recv_multipart(
+        self, flags: int = 0, *, copy: Literal[False], track: bool = False
+    ) -> Awaitable[List[_zmq.Frame]]:  # type: ignore
+        ...
+
+    @overload
+    def recv_multipart(
+        self, flags: int = 0, copy: bool = True, track: bool = False
+    ) -> Awaitable[Union[List[bytes], List[_zmq.Frame]]]:
+        ...
+
+    def recv_multipart(
+        self, flags: int = 0, copy: bool = True, track: bool = False
+    ) -> Awaitable[Union[List[bytes], List[_zmq.Frame]]]:
         """Receive a complete multipart zmq message.
 
         Returns a Future whose result will be a multipart message.
@@ -187,7 +298,9 @@ class _AsyncSocket(_zmq.Socket):
             'recv_multipart', dict(flags=flags, copy=copy, track=track)
         )
 
-    def recv(self, flags=0, copy=True, track=False):
+    def recv(  # type: ignore
+        self, flags: int = 0, copy: bool = True, track: bool = False
+    ) -> Awaitable[Union[bytes, _zmq.Frame]]:
         """Receive a single zmq frame.
 
         Returns a Future, whose result will be the received frame.
@@ -196,7 +309,9 @@ class _AsyncSocket(_zmq.Socket):
         """
         return self._add_recv_event('recv', dict(flags=flags, copy=copy, track=track))
 
-    def send_multipart(self, msg, flags=0, copy=True, track=False, **kwargs):
+    def send_multipart(  # type: ignore
+        self, msg_parts: Any, flags: int = 0, copy: bool = True, track=False, **kwargs
+    ) -> Awaitable[Optional[_zmq.MessageTracker]]:
         """Send a complete multipart zmq message.
 
         Returns a Future that resolves when sending is complete.
@@ -204,9 +319,16 @@ class _AsyncSocket(_zmq.Socket):
         kwargs['flags'] = flags
         kwargs['copy'] = copy
         kwargs['track'] = track
-        return self._add_send_event('send_multipart', msg=msg, kwargs=kwargs)
+        return self._add_send_event('send_multipart', msg=msg_parts, kwargs=kwargs)
 
-    def send(self, msg, flags=0, copy=True, track=False, **kwargs):
+    def send(  # type: ignore
+        self,
+        data: Any,
+        flags: int = 0,
+        copy: bool = True,
+        track: bool = False,
+        **kwargs: Any,
+    ) -> Awaitable[Optional[_zmq.MessageTracker]]:
         """Send a single zmq frame.
 
         Returns a Future that resolves when sending is complete.
@@ -217,7 +339,7 @@ class _AsyncSocket(_zmq.Socket):
         kwargs['copy'] = copy
         kwargs['track'] = track
         kwargs.update(dict(flags=flags, copy=copy, track=track))
-        return self._add_send_event('send', msg=msg, kwargs=kwargs)
+        return self._add_send_event('send', msg=data, kwargs=kwargs)
 
     def _deserialize(self, recvd, load):
         """Deserialize with Futures"""
@@ -251,7 +373,7 @@ class _AsyncSocket(_zmq.Socket):
 
         return f
 
-    def poll(self, timeout=None, flags=_zmq.POLLIN):
+    def poll(self, timeout=None, flags=_zmq.POLLIN) -> Awaitable[int]:  # type: ignore
         """poll the socket for events
 
         returns a Future for the poll results.
@@ -262,7 +384,7 @@ class _AsyncSocket(_zmq.Socket):
 
         p = self._poller_class()
         p.register(self, flags)
-        f = p.poll(timeout)
+        f = cast(Future, p.poll(timeout))
 
         future = self._Future()
 
@@ -289,6 +411,13 @@ class _AsyncSocket(_zmq.Socket):
             f.add_done_callback(unwrap_result)
         return future
 
+    # overrides only necessary for updated types
+    def recv_string(self, *args, **kwargs) -> Awaitable[str]:  # type: ignore
+        return super().recv_string(*args, **kwargs)  # type: ignore
+
+    def send_string(self, s: str, flags: int = 0, encoding: str = 'utf-8') -> Awaitable[None]:  # type: ignore
+        return super().send_string(s, flags=flags, encoding=encoding)  # type: ignore
+
     def _add_timeout(self, future, timeout):
         """Add a timeout for a send or recv Future"""
 
@@ -300,7 +429,7 @@ class _AsyncSocket(_zmq.Socket):
             # raise EAGAIN
             future.set_exception(_zmq.Again())
 
-        self._call_later(timeout, future_timeout)
+        return self._call_later(timeout, future_timeout)
 
     def _call_later(self, delay, callback):
         """Schedule a function to be called later
@@ -310,7 +439,7 @@ class _AsyncSocket(_zmq.Socket):
         Tornado and asyncio happen to both have ioloop.call_later
         with the same signature.
         """
-        self.io_loop.call_later(delay, callback)
+        return self._get_loop().call_later(delay, callback)
 
     @staticmethod
     def _remove_finished_future(future, event_list):
@@ -319,8 +448,8 @@ class _AsyncSocket(_zmq.Socket):
         Avoids delaying cleanup until the next send/recv event,
         which may never come.
         """
-        for f_idx, (f, kind, kwargs, _) in enumerate(event_list):
-            if f is future:
+        for f_idx, event in enumerate(event_list):
+            if event.future is future:
                 break
         else:
             return
@@ -342,20 +471,21 @@ class _AsyncSocket(_zmq.Socket):
                 f.set_result(r)
             return f
 
+        timer = _NoTimer
+        if hasattr(_zmq, 'RCVTIMEO'):
+            timeout_ms = self._shadow_sock.rcvtimeo
+            if timeout_ms >= 0:
+                timer = self._add_timeout(f, timeout_ms * 1e-3)
+
         # we add it to the list of futures before we add the timeout as the
         # timeout will remove the future from recv_futures to avoid leaks
-        self._recv_futures.append(_FutureEvent(f, kind, kwargs, msg=None))
+        self._recv_futures.append(_FutureEvent(f, kind, kwargs, msg=None, timer=timer))
 
         # Don't let the Future sit in _recv_events after it's done
         f.add_done_callback(
             lambda f: self._remove_finished_future(f, self._recv_futures)
         )
 
-        if hasattr(_zmq, 'RCVTIMEO'):
-            timeout_ms = self._shadow_sock.rcvtimeo
-            if timeout_ms >= 0:
-                self._add_timeout(f, timeout_ms * 1e-3)
-
         if self._shadow_sock.get(EVENTS) & POLLIN:
             # recv immediately, if we can
             self._handle_recv()
@@ -400,19 +530,22 @@ class _AsyncSocket(_zmq.Socket):
                     self._schedule_remaining_events()
                 return f
 
+        timer = _NoTimer
+        if hasattr(_zmq, 'SNDTIMEO'):
+            timeout_ms = self._shadow_sock.get(_zmq.SNDTIMEO)
+            if timeout_ms >= 0:
+                timer = self._add_timeout(f, timeout_ms * 1e-3)
+
         # we add it to the list of futures before we add the timeout as the
         # timeout will remove the future from recv_futures to avoid leaks
-        self._send_futures.append(_FutureEvent(f, kind, kwargs=kwargs, msg=msg))
+        self._send_futures.append(
+            _FutureEvent(f, kind, kwargs=kwargs, msg=msg, timer=timer)
+        )
         # Don't let the Future sit in _send_futures after it's done
         f.add_done_callback(
             lambda f: self._remove_finished_future(f, self._send_futures)
         )
 
-        if hasattr(_zmq, 'SNDTIMEO'):
-            timeout_ms = self._shadow_sock.get(_zmq.SNDTIMEO)
-            if timeout_ms >= 0:
-                self._add_timeout(f, timeout_ms * 1e-3)
-
         self._add_io_state(POLLOUT)
         return f
 
@@ -423,7 +556,7 @@ class _AsyncSocket(_zmq.Socket):
             return
         f = None
         while self._recv_futures:
-            f, kind, kwargs, _ = self._recv_futures.popleft()
+            f, kind, kwargs, _, timer = self._recv_futures.popleft()
             # skip any cancelled futures
             if f.done():
                 f = None
@@ -436,6 +569,8 @@ class _AsyncSocket(_zmq.Socket):
         if f is None:
             return
 
+        timer.cancel()
+
         if kind == 'poll':
             # on poll event, just signal ready, nothing else.
             f.set_result(None)
@@ -461,7 +596,7 @@ class _AsyncSocket(_zmq.Socket):
             return
         f = None
         while self._send_futures:
-            f, kind, kwargs, msg = self._send_futures.popleft()
+            f, kind, kwargs, msg, timer = self._send_futures.popleft()
             # skip any cancelled futures
             if f.done():
                 f = None
@@ -474,6 +609,8 @@ class _AsyncSocket(_zmq.Socket):
         if f is None:
             return
 
+        timer.cancel()
+
         if kind == 'poll':
             # on poll event, just signal ready, nothing else.
             f.set_result(None)
@@ -536,11 +673,17 @@ class _AsyncSocket(_zmq.Socket):
 
         zmq FD is always read-only.
         """
+        # ensure loop is registered and init_io has been called
+        # if there are any events to watch for
+        if state:
+            self._get_loop()
         self._schedule_remaining_events()
 
-    def _init_io_state(self):
+    def _init_io_state(self, loop=None):
         """initialize the ioloop event handler"""
-        self.io_loop.add_handler(self._shadow_sock, self._handle_events, self._READ)
+        if loop is None:
+            loop = self._get_loop()
+        loop.add_handler(self._shadow_sock, self._handle_events, self._READ)
         self._call_later(0, self._handle_events)
 
     def _clear_io_state(self):
@@ -551,4 +694,5 @@ class _AsyncSocket(_zmq.Socket):
         fd = self._shadow_sock
         if self._shadow_sock.closed:
             fd = self._fd
-        self.io_loop.remove_handler(fd)
+        if self._current_loop is not None:
+            self._current_loop.remove_handler(fd)
